# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2021-05-24 12:51
from hanlp_common.constant import HANLP_URL

STS_ELECTRA_BASE_ZH = HANLP_URL + 'sts/sts_electra_base_zh_20210524_130550.zip'
'A naive regression model trained on concatenated STS corpora.'

# Will be filled up during runtime
ALL = {}
