# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-28 22:19

_CTB6_CWS_HOME = 'http://file.hankcs.com/corpus/ctb6_cws.zip'

CTB6_CWS_TRAIN = _CTB6_CWS_HOME + '#train.txt'
'''CTB6 training set.'''
CTB6_CWS_DEV = _CTB6_CWS_HOME + '#dev.txt'
'''CTB6 dev set.'''
CTB6_CWS_TEST = _CTB6_CWS_HOME + '#test.txt'
'''CTB6 test set.'''
